// GLOBAL PROTOCOLS
// __CTX__START        Start a background process
// __CTX__CONNECT      Start a background process
// __CTX__RESTART      Restart a service background process
// __CTX__SERVICE_#ID          Protocol layer to communicate with a service.
// __CTX__REQUIRE              Require a module

var io = require('socket.io-client');
var toArray = require('to-array');
var Emitter = require('component-emitter');
var async = require('async');
var manager = new io.Manager('', {autoConnect:false});
var localhost = manager.socket('/__CTX__');

// This causes the localhost to emit on itself rather than trying to
// communicate over a socket.
localhost.emit = (new Emitter()).emit.bind(localhost);

var origin = '';
Object.defineProperty(CTX.config, 'origin', {
    set : function(new_origin){
        if(typeof new_origin != 'string') return;

        CTX._origins[new_origin] = CTX._origins[origin];
        delete CTX._origins[origin];

        origin = new_origin;
    },
    get : function(){
        return origin;
    }
});

Object.defineProperty(CTX, '_localhost', {
    set : function(socket){
        CTX._origins[origin] = socket;
    },
    get : function(){
        return CTX._origins[origin];
    }
});

CTX._localhost = localhost;
localhost.on('__CTX__REQUIRE', function(requests, done){
    CTX.Module._fetch(requests, done);
});

localhost.on('__CTX__START', function(options, done){
    if(options.unique !== true){
        var id = CTX._generateId(options.path);
        if(CTX.Service.running[id]){
            done(undefined, CTX.Service.running[id]);
            return;
        }
    }

    fetchService(options, function(err, _package){
        if(err){
            done(err);
        }else{
            // Because localhost is not a socket the service will
            // replace the provided socket with a blank placeholder socket.
            // This is the desired behaviour as there is no foreground
            // connected yet, so we want a disconnected socket so that the
            // events will buffer until an active socket is connected.

            var id = options.unique === true ?  CTX._generateId() : CTX._generateId(options.path);

            try{
                var service = new CTX.Service(_package, options);
                service._id = id;

                if(options.data){
                    service.data = options.data;
                }
            }catch(err){
                done(err);
                return;
            }

            done(undefined, service);
        }
    });

    function fetchService(options, done){
        var _package = CTX.Service.cache[options.path];
        if(_package && CTX.config.cache !== false){
            done(null, _package);
        }else{
            // Need to remove any protocols and add them back later.
            var base_path = CTX._path.join(CTX.config.path, options.path);
            CTX._fs.readFile(base_path+'/service.ctx.json', function(error, file){
                // Starts building compiling assets package
                if(error){
                    done(error);
                    return;
                }

                var _descriptor = {};
                try{
                    _descriptor = JSON.parse(file);
                }catch(error){
                    done(error);
                    return;
                }

                var _package = {
                    path    : options.path,
                    controllers : {},
                    _descriptor : _descriptor
                }

                _descriptor.controllers = _descriptor.controllers instanceof Object ? _descriptor.controllers : {};

                async.parallel({
                    bg : function(cb){
                        if(_descriptor.controllers.background){
                            CTX._fs.readFile(CTX._path.join(base_path, _descriptor.controllers.background), function(error, file){
                                if(error){
                                    console.error('CTX: Failed to read controller for "'+_package.path+'".');
                                    if(CTX.config.verbose) console.error(error);
                                    else console.error(error.message);
                                }else{
                                    try{
                                        if(!/sourceURL=/g.test(file)){
                                            file += '\n//# sourceURL='+_package.path
                                        }

                                        _package.controllers.background = file;
                                    }catch(error){
                                        console.error('CTX: Failed to evaluate controller for "'+_package.path+'".');
                                        if(CTX.config.verbose) console.error(error);
                                        else console.error(error.message);
                                    }
                                }

                                cb();
                            });
                        }else{
                            cb();
                        }
                    },
                    fg : function(cb){
                        if(_descriptor.controllers.foreground){
                            CTX._fs.readFile(CTX._path.join(base_path, _descriptor.controllers.foreground), function(error, file){
                                if(error){
                                    console.error('CTX: Failed to read controller for "'+_package.path+'".');
                                    if(CTX.config.verbose) console.error(error);
                                    else console.error(error.message);
                                }else{
                                    if(!/sourceURL=/g.test(file)){
                                        file += '\n//# sourceURL='+_package.path;
                                    }
                                    _package.controllers.foreground = file;
                                }

                                cb();
                            });
                        }else{
                            cb();
                        }
                    }
                }, function(error, data){
                    if(error){
                        done(error);
                    }else{
                        try{
                            _package.controller = new Function('require', 'global', 'process',_package.controllers.background);
                        }catch(error){
                            console.error('CTX: Failed to evaluate controller for "'+_package.path+'".');
                            if(CTX.config.verbose) console.error(error);
                            else console.error(error.message);
                        }

                        CTX.Service.cache[_package.path] = _package;
                        done(undefined,  _package);
                    }
                });
            });
        }
    }
});

localhost.connect = function(socket){
    // Bridges socket protocol events from a socket to localhost
    // for CTX based functions.

    socket.on('__CTX__CONNECT', function(_id){
        try{
            CTX.Service.running[_id]._socket = socket;
        }catch(error){}
    });

    socket.on('__CTX__START', function(options, done){
        options = options instanceof Object ? options : {};
        done = done instanceof Function ? done : function(){};
        options.new = true; // Force new start from clients.

        localhost.emit('__CTX__START', options, function(err, _package){
            if(err){
                done(err);
            }else{
                try{
                    var service = new CTX.Service(_package);
                }catch(err){
                    done(err);
                    return;
                }
                done(undefined, service);
            }
        });
    });


    socket.on('__CTX__REQUIRE', function(requests, done){
        CTX.Module._fetch(requests, done);
    });
}



module.exports = localhost;
